/*
 * linux/fs/xenix/symlink.c
 *
 * Copyright (C) 1991, 1992  Linus Torvalds
 *
 * Taken from minix/symlink.c, modified by Doug Evans, dje@sspiff.uucp, 92Jun04.
 *
 * This file contains xenix symlink handling code. While xenix filesystems
 * don't support symbolic links, one can make a non-upward compatible extension
 * and add them if one wants to.
 *
 * I've added support for an upward compatible implementation of symbolic
 * links. It's quite kludgey, but it doesn't send fsck into fits like a real
 * implementation would. Regular files that have mode 01000 (IE: only the
 * sticky bit set) are treated as symbolic links. dje@cygnus.com, 92Nov02.
 * This translation is made when the inode is read/written from/to disk so the
 * changes are minimal.
 */

#include <asm/segment.h>

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/xenix_fs.h>
#include <linux/stat.h>

static int xenix_readlink(struct inode *, char *, int);
static int xenix_follow_link(struct inode *, struct inode *, int, int, struct inode **);

/*
 * symlinks can't do much...
 */

struct inode_operations xenix_symlink_inode_operations = {
	NULL,			/* no file-operations */
	NULL,			/* create */
	NULL,			/* lookup */
	NULL,			/* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	xenix_readlink,		/* readlink */
	xenix_follow_link,	/* follow_link */
	NULL,			/* bmap */
	NULL,			/* truncate */
	NULL			/* permission */
};

static int
xenix_follow_link(struct inode * dir, struct inode * inode,
	int flag, int mode, struct inode ** res_inode)
{
	int error;
	struct buffer_head * bh;

	*res_inode = NULL;
	if (!dir) {
		dir = current->root;
		dir->i_count++;
	}
	if (!inode) {
		iput(dir);
		return -ENOENT;
	}
	if (!S_ISLNK(inode->i_mode)) {
		iput(dir);
		*res_inode = inode;
		return 0;
	}
	if (current->link_count > 5) {
		iput(inode);
		iput(dir);
		return -ELOOP;
	}
	if (!(bh = xenix_bread(inode, 0, 0))) {
		iput(inode);
		iput(dir);
		return -EIO;
	}
	iput(inode);
	current->link_count++;
	error = open_namei(bh->b_data, flag, mode, res_inode, dir);
	current->link_count--;
	brelse(bh);
	return error;
}

static int
xenix_readlink(struct inode * inode, char * buffer, int buflen)
{
	struct buffer_head * bh;
	int i;
	char c;

	if (!S_ISLNK(inode->i_mode)) {
		iput(inode);
		return -EINVAL;
	}
	if (buflen > 1023)
		buflen = 1023;
	bh = xenix_bread(inode, 0, 0);
	iput(inode);
	if (!bh)
		return 0;
	i = 0;
	while (i < buflen && (c = bh->b_data[i])) {
		i++;
		put_fs_byte(c, buffer++);
	}
	brelse(bh);
	return i;
}
